# Copyright (c) 2020 Jisang Yoon
# All rights reserved.
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.
from cuhnsw.pyhnsw import CuHNSW



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x0c\x63onfig.proto\"\x8e\x03\n\x0b\x43onfigProto\x12\x11\n\x04seed\x18\x01 \x01(\x05:\x03\x37\x37\x37\x12\x16\n\x0b\x63_log_level\x18\x03 \x01(\x05:\x01\x32\x12\x17\n\x0cpy_log_level\x18\x04 \x01(\x05:\x01\x32\x12\x11\n\x05max_m\x18\x05 \x01(\x05:\x02\x31\x32\x12\x12\n\x06max_m0\x18\x06 \x01(\x05:\x02\x32\x34\x12\x1c\n\x0f\x65\x66_construction\x18\x07 \x01(\x05:\x03\x31\x35\x30\x12\x12\n\nlevel_mult\x18\t \x01(\x01\x12\x14\n\x0csave_remains\x18\n \x01(\x08\x12\x19\n\rhyper_threads\x18\x0b \x01(\x01:\x02\x31\x30\x12\x15\n\tblock_dim\x18\x0c \x01(\x05:\x02\x33\x32\x12\x16\n\tdist_type\x18\r \x01(\t:\x03\x64ot\x12\x1a\n\x12visited_table_size\x18\x11 \x01(\x05\x12\x1f\n\x11visited_list_size\x18\x0e \x01(\x05:\x04\x38\x31\x39\x32\x12\x11\n\x03nrz\x18\x0f \x01(\x08:\x04true\x12\x14\n\x0creverse_cand\x18\x10 \x01(\x08\x12\x1c\n\x0eheuristic_coef\x18\x12 \x01(\x01:\x04\x30.25'
)




_CONFIGPROTO = _descriptor.Descriptor(
  name='ConfigProto',
  full_name='ConfigProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='seed', full_name='ConfigProto.seed', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=777,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='c_log_level', full_name='ConfigProto.c_log_level', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='py_log_level', full_name='ConfigProto.py_log_level', index=2,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_m', full_name='ConfigProto.max_m', index=3,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=12,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_m0', full_name='ConfigProto.max_m0', index=4,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=24,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ef_construction', full_name='ConfigProto.ef_construction', index=5,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=150,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level_mult', full_name='ConfigProto.level_mult', index=6,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_remains', full_name='ConfigProto.save_remains', index=7,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hyper_threads', full_name='ConfigProto.hyper_threads', index=8,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=float(10),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_dim', full_name='ConfigProto.block_dim', index=9,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=32,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dist_type', full_name='ConfigProto.dist_type', index=10,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"dot".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visited_table_size', full_name='ConfigProto.visited_table_size', index=11,
      number=17, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visited_list_size', full_name='ConfigProto.visited_list_size', index=12,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=8192,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nrz', full_name='ConfigProto.nrz', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reverse_cand', full_name='ConfigProto.reverse_cand', index=14,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='heuristic_coef', full_name='ConfigProto.heuristic_coef', index=15,
      number=18, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=float(0.25),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=415,
)

DESCRIPTOR.message_types_by_name['ConfigProto'] = _CONFIGPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ConfigProto = _reflection.GeneratedProtocolMessageType('ConfigProto', (_message.Message,), {
  'DESCRIPTOR' : _CONFIGPROTO,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:ConfigProto)
  })
_sym_db.RegisterMessage(ConfigProto)


# @@protoc_insertion_point(module_scope)



# Copyright (c) 2020 Jisang Yoon
# All rights reserved.
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.
import re
import json
import logging
import logging.handlers

import jsmin
from google.protobuf.json_format import Parse, MessageToDict
from cuhnsw.config_pb2 import ConfigProto

# get_logger and Option refer to
# https://github.com/kakao/buffalo/blob/
# 5f571c2c7d8227e6625c6e538da929e4db11b66d/buffalo/misc/aux.py
def get_logger(name=__file__, level=2):
  if level == 1:
    level = logging.WARNING
  elif level == 2:
    level = logging.INFO
  elif level == 3:
    level = logging.DEBUG
  logger = logging.getLogger(name)
  if logger.handlers:
    return logger
  logger.setLevel(level)
  sh0 = logging.StreamHandler()
  sh0.setLevel(level)
  formatter = logging.Formatter('[%(levelname)-8s] %(asctime)s '
                                '[%(filename)s] [%(funcName)s:%(lineno)d]'
                                '%(message)s', '%Y-%m-%d %H:%M:%S')
  sh0.setFormatter(formatter)
  logger.addHandler(sh0)
  return logger

# This function helps you to read non-standard json strings.
# - Handles json string with c++ style inline comments
# - Handles json string with trailing commas.
def load_json_string(cont):
  # (1) Removes comment.
  #     Refer to https://plus.google.com/+DouglasCrockfordEsq/posts/RK8qyGVaGSr
  cont = jsmin.jsmin(cont)

  # (2) Removes trailing comma.
  cont = re.sub(",[ \t\r\n]*}", "}", cont)
  cont = re.sub(",[ \t\r\n]*" + r"\]", "]", cont)

  return json.loads(cont)


# function read json file from filename
def load_json_file(fname):
  with open(fname, "r") as fin:
    ret = load_json_string(fin.read())
  return ret

# use protobuf to restrict field and types
def get_opt_as_proto(raw, proto_type=ConfigProto):
  proto = proto_type()
  # convert raw to proto
  Parse(json.dumps(Option(raw)), proto)
  err = []
  assert proto.IsInitialized(err), \
    f"some required fields are missing in proto {err}\n {proto}"
  return proto

def proto_to_dict(proto):
  return MessageToDict(proto, \
    including_default_value_fields=True, \
    preserving_proto_field_name=True)

def copy_proto(proto):
  newproto = type(proto)()
  Parse(json.dumps(proto_to_dict(proto)), newproto)
  return newproto

class Option(dict):
  def __init__(self, *args, **kwargs):
    args = [arg if isinstance(arg, dict)
            else load_json_file(arg) for arg in args]
    super().__init__(*args, **kwargs)
    for arg in args:
      if isinstance(arg, dict):
        for k, val in arg.items():
          if isinstance(val, dict):
            self[k] = Option(val)
          else:
            self[k] = val
    if kwargs:
      for k, val in kwargs.items():
        if isinstance(val, dict):
          self[k] = Option(val)
        else:
          self[k] = val

  def __getattr__(self, attr):
    return self.get(attr)

  def __setattr__(self, key, value):
    self.__setitem__(key, value)

  def __setitem__(self, key, value):
    super().__setitem__(key, value)
    self.__dict__.update({key: value})

  def __delattr__(self, item):
    self.__delitem__(item)

  def __delitem__(self, key):
    super().__delitem__(key)
    del self.__dict__[key]

  def __getstate__(self):
    return vars(self)

  def __setstate__(self, state):
    vars(self).update(state)



# -*- coding: utf-8 -*-
from __future__ import division
import pytest
import sys

import env  # noqa: F401

from pybind11_tests import pytypes as m
from pybind11_tests import debug_enabled


def test_int(doc):
    assert doc(m.get_int) == "get_int() -> int"


def test_iterator(doc):
    assert doc(m.get_iterator) == "get_iterator() -> Iterator"


def test_iterable(doc):
    assert doc(m.get_iterable) == "get_iterable() -> Iterable"


def test_list(capture, doc):
    with capture:
        lst = m.get_list()
        assert lst == ["inserted-0", "overwritten", "inserted-2"]

        lst.append("value2")
        m.print_list(lst)
    assert (
        capture.unordered
        == """
        Entry at position 0: value
        list item 0: inserted-0
        list item 1: overwritten
        list item 2: inserted-2
        list item 3: value2
    """
    )

    assert doc(m.get_list) == "get_list() -> list"
    assert doc(m.print_list) == "print_list(arg0: list) -> None"


def test_none(capture, doc):
    assert doc(m.get_none) == "get_none() -> None"
    assert doc(m.print_none) == "print_none(arg0: None) -> None"


def test_set(capture, doc):
    s = m.get_set()
    assert s == {"key1", "key2", "key3"}

    with capture:
        s.add("key4")
        m.print_set(s)
    assert (
        capture.unordered
        == """
        key: key1
        key: key2
        key: key3
        key: key4
    """
    )

    assert not m.set_contains(set([]), 42)
    assert m.set_contains({42}, 42)
    assert m.set_contains({"foo"}, "foo")

    assert doc(m.get_list) == "get_list() -> list"
    assert doc(m.print_list) == "print_list(arg0: list) -> None"


def test_dict(capture, doc):
    d = m.get_dict()
    assert d == {"key": "value"}

    with capture:
        d["key2"] = "value2"
        m.print_dict(d)
    assert (
        capture.unordered
        == """
        key: key, value=value
        key: key2, value=value2
    """
    )

    assert not m.dict_contains({}, 42)
    assert m.dict_contains({42: None}, 42)
    assert m.dict_contains({"foo": None}, "foo")

    assert doc(m.get_dict) == "get_dict() -> dict"
    assert doc(m.print_dict) == "print_dict(arg0: dict) -> None"

    assert m.dict_keyword_constructor() == {"x": 1, "y": 2, "z": 3}


def test_str(doc):
    assert m.str_from_string().encode().decode() == "baz"
    assert m.str_from_bytes().encode().decode() == "boo"

    assert doc(m.str_from_bytes) == "str_from_bytes() -> str"

    class A(object):
        def __str__(self):
            return "this is a str"

        def __repr__(self):
            return "this is a repr"

    assert m.str_from_object(A()) == "this is a str"
    assert m.repr_from_object(A()) == "this is a repr"
    assert m.str_from_handle(A()) == "this is a str"

    s1, s2 = m.str_format()
    assert s1 == "1 + 2 = 3"
    assert s1 == s2

    malformed_utf8 = b"\x80"
    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE"):
        assert m.str_from_object(malformed_utf8) is malformed_utf8
    elif env.PY2:
        with pytest.raises(UnicodeDecodeError):
            m.str_from_object(malformed_utf8)
    else:
        assert m.str_from_object(malformed_utf8) == "b'\\x80'"
    if env.PY2:
        with pytest.raises(UnicodeDecodeError):
            m.str_from_handle(malformed_utf8)
    else:
        assert m.str_from_handle(malformed_utf8) == "b'\\x80'"


def test_bytes(doc):
    assert m.bytes_from_string().decode() == "foo"
    assert m.bytes_from_str().decode() == "bar"

    assert doc(m.bytes_from_str) == "bytes_from_str() -> {}".format(
        "str" if env.PY2 else "bytes"
    )


def test_capsule(capture):
    pytest.gc_collect()
    with capture:
        a = m.return_capsule_with_destructor()
        del a
        pytest.gc_collect()
    assert (
        capture.unordered
        == """
        creating capsule
        destructing capsule
    """
    )

    with capture:
        a = m.return_capsule_with_destructor_2()
        del a
        pytest.gc_collect()
    assert (
        capture.unordered
        == """
        creating capsule
        destructing capsule: 1234
    """
    )

    with capture:
        a = m.return_capsule_with_name_and_destructor()
        del a
        pytest.gc_collect()
    assert (
        capture.unordered
        == """
        created capsule (1234, 'pointer type description')
        destructing capsule (1234, 'pointer type description')
    """
    )


def test_accessors():
    class SubTestObject:
        attr_obj = 1
        attr_char = 2

    class TestObject:
        basic_attr = 1
        begin_end = [1, 2, 3]
        d = {"operator[object]": 1, "operator[char *]": 2}
        sub = SubTestObject()

        def func(self, x, *args):
            return self.basic_attr + x + sum(args)

    d = m.accessor_api(TestObject())
    assert d["basic_attr"] == 1
    assert d["begin_end"] == [1, 2, 3]
    assert d["operator[object]"] == 1
    assert d["operator[char *]"] == 2
    assert d["attr(object)"] == 1
    assert d["attr(char *)"] == 2
    assert d["missing_attr_ptr"] == "raised"
    assert d["missing_attr_chain"] == "raised"
    assert d["is_none"] is False
    assert d["operator()"] == 2
    assert d["operator*"] == 7
    assert d["implicit_list"] == [1, 2, 3]
    assert all(x in TestObject.__dict__ for x in d["implicit_dict"])

    assert m.tuple_accessor(tuple()) == (0, 1, 2)

    d = m.accessor_assignment()
    assert d["get"] == 0
    assert d["deferred_get"] == 0
    assert d["set"] == 1
    assert d["deferred_set"] == 1
    assert d["var"] == 99


def test_constructors():
    """C++ default and converting constructors are equivalent to type calls in Python"""
    types = [bytes, str, bool, int, float, tuple, list, dict, set]
    expected = {t.__name__: t() for t in types}
    if env.PY2:
        # Note that bytes.__name__ == 'str' in Python 2.
        # pybind11::str is unicode even under Python 2.
        expected["bytes"] = bytes()
        expected["str"] = unicode()  # noqa: F821
    assert m.default_constructors() == expected

    data = {
        bytes: b"41",  # Currently no supported or working conversions.
        str: 42,
        bool: "Not empty",
        int: "42",
        float: "+1e3",
        tuple: range(3),
        list: range(3),
        dict: [("two", 2), ("one", 1), ("three", 3)],
        set: [4, 4, 5, 6, 6, 6],
        memoryview: b"abc",
    }
    inputs = {k.__name__: v for k, v in data.items()}
    expected = {k.__name__: k(v) for k, v in data.items()}
    if env.PY2:  # Similar to the above. See comments above.
        inputs["bytes"] = b"41"
        inputs["str"] = 42
        expected["bytes"] = b"41"
        expected["str"] = u"42"

    assert m.converting_constructors(inputs) == expected
    assert m.cast_functions(inputs) == expected

    # Converting constructors and cast functions should just reference rather
    # than copy when no conversion is needed:
    noconv1 = m.converting_constructors(expected)
    for k in noconv1:
        assert noconv1[k] is expected[k]

    noconv2 = m.cast_functions(expected)
    for k in noconv2:
        assert noconv2[k] is expected[k]


def test_non_converting_constructors():
    non_converting_test_cases = [
        ("bytes", range(10)),
        ("none", 42),
        ("ellipsis", 42),
        ("type", 42),
    ]
    for t, v in non_converting_test_cases:
        for move in [True, False]:
            with pytest.raises(TypeError) as excinfo:
                m.nonconverting_constructor(t, v, move)
            expected_error = "Object of type '{}' is not an instance of '{}'".format(
                type(v).__name__, t
            )
            assert str(excinfo.value) == expected_error


def test_pybind11_str_raw_str():
    # specifically to exercise pybind11::str::raw_str
    cvt = m.convert_to_pybind11_str
    assert cvt(u"Str") == u"Str"
    assert cvt(b"Bytes") == u"Bytes" if env.PY2 else "b'Bytes'"
    assert cvt(None) == u"None"
    assert cvt(False) == u"False"
    assert cvt(True) == u"True"
    assert cvt(42) == u"42"
    assert cvt(2 ** 65) == u"36893488147419103232"
    assert cvt(-1.50) == u"-1.5"
    assert cvt(()) == u"()"
    assert cvt((18,)) == u"(18,)"
    assert cvt([]) == u"[]"
    assert cvt([28]) == u"[28]"
    assert cvt({}) == u"{}"
    assert cvt({3: 4}) == u"{3: 4}"
    assert cvt(set()) == u"set([])" if env.PY2 else "set()"
    assert cvt({3, 3}) == u"set([3])" if env.PY2 else "{3}"

    valid_orig = u"Ǳ"
    valid_utf8 = valid_orig.encode("utf-8")
    valid_cvt = cvt(valid_utf8)
    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE"):
        assert valid_cvt is valid_utf8
    else:
        assert type(valid_cvt) is unicode if env.PY2 else str  # noqa: F821
        if env.PY2:
            assert valid_cvt == valid_orig
        else:
            assert valid_cvt == "b'\\xc7\\xb1'"

    malformed_utf8 = b"\x80"
    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE"):
        assert cvt(malformed_utf8) is malformed_utf8
    else:
        if env.PY2:
            with pytest.raises(UnicodeDecodeError):
                cvt(malformed_utf8)
        else:
            malformed_cvt = cvt(malformed_utf8)
            assert type(malformed_cvt) is str
            assert malformed_cvt == "b'\\x80'"


def test_implicit_casting():
    """Tests implicit casting when assigning or appending to dicts and lists."""
    z = m.get_implicit_casting()
    assert z["d"] == {
        "char*_i1": "abc",
        "char*_i2": "abc",
        "char*_e": "abc",
        "char*_p": "abc",
        "str_i1": "str",
        "str_i2": "str1",
        "str_e": "str2",
        "str_p": "str3",
        "int_i1": 42,
        "int_i2": 42,
        "int_e": 43,
        "int_p": 44,
    }
    assert z["l"] == [3, 6, 9, 12, 15]


def test_print(capture):
    with capture:
        m.print_function()
    assert (
        capture
        == """
        Hello, World!
        1 2.0 three True -- multiple args
        *args-and-a-custom-separator
        no new line here -- next print
        flush
        py::print + str.format = this
    """
    )
    assert capture.stderr == "this goes to stderr"

    with pytest.raises(RuntimeError) as excinfo:
        m.print_failure()
    assert str(excinfo.value) == "make_tuple(): unable to convert " + (
        "argument of type 'UnregisteredType' to Python object"
        if debug_enabled
        else "arguments to Python object (compile in debug mode for details)"
    )


def test_hash():
    class Hashable(object):
        def __init__(self, value):
            self.value = value

        def __hash__(self):
            return self.value

    class Unhashable(object):
        __hash__ = None

    assert m.hash_function(Hashable(42)) == 42
    with pytest.raises(TypeError):
        m.hash_function(Unhashable())


def test_number_protocol():
    for a, b in [(1, 1), (3, 5)]:
        li = [
            a == b,
            a != b,
            a < b,
            a <= b,
            a > b,
            a >= b,
            a + b,
            a - b,
            a * b,
            a / b,
            a | b,
            a & b,
            a ^ b,
            a >> b,
            a << b,
        ]
        assert m.test_number_protocol(a, b) == li


def test_list_slicing():
    li = list(range(100))
    assert li[::2] == m.test_list_slicing(li)


def test_issue2361():
    # See issue #2361
    assert m.issue2361_str_implicit_copy_none() == "None"
    with pytest.raises(TypeError) as excinfo:
        assert m.issue2361_dict_implicit_copy_none()
    assert "'NoneType' object is not iterable" in str(excinfo.value)


@pytest.mark.parametrize(
    "method, args, fmt, expected_view",
    [
        (m.test_memoryview_object, (b"red",), "B", b"red"),
        (m.test_memoryview_buffer_info, (b"green",), "B", b"green"),
        (m.test_memoryview_from_buffer, (False,), "h", [3, 1, 4, 1, 5]),
        (m.test_memoryview_from_buffer, (True,), "H", [2, 7, 1, 8]),
        (m.test_memoryview_from_buffer_nativeformat, (), "@i", [4, 7, 5]),
    ],
)
def test_memoryview(method, args, fmt, expected_view):
    view = method(*args)
    assert isinstance(view, memoryview)
    assert view.format == fmt
    if isinstance(expected_view, bytes) or not env.PY2:
        view_as_list = list(view)
    else:
        # Using max to pick non-zero byte (big-endian vs little-endian).
        view_as_list = [max([ord(c) for c in s]) for s in view]
    assert view_as_list == list(expected_view)


@pytest.mark.xfail("env.PYPY", reason="getrefcount is not available")
@pytest.mark.parametrize(
    "method",
    [
        m.test_memoryview_object,
        m.test_memoryview_buffer_info,
    ],
)
def test_memoryview_refcount(method):
    buf = b"\x0a\x0b\x0c\x0d"
    ref_before = sys.getrefcount(buf)
    view = method(buf)
    ref_after = sys.getrefcount(buf)
    assert ref_before < ref_after
    assert list(view) == list(buf)


def test_memoryview_from_buffer_empty_shape():
    view = m.test_memoryview_from_buffer_empty_shape()
    assert isinstance(view, memoryview)
    assert view.format == "B"
    if env.PY2:
        # Python 2 behavior is weird, but Python 3 (the future) is fine.
        # PyPy3 has <memoryview, while CPython 2 has <memory
        assert bytes(view).startswith(b"<memory")
    else:
        assert bytes(view) == b""


def test_test_memoryview_from_buffer_invalid_strides():
    with pytest.raises(RuntimeError):
        m.test_memoryview_from_buffer_invalid_strides()


def test_test_memoryview_from_buffer_nullptr():
    if env.PY2:
        m.test_memoryview_from_buffer_nullptr()
    else:
        with pytest.raises(ValueError):
            m.test_memoryview_from_buffer_nullptr()


@pytest.mark.skipif("env.PY2")
def test_memoryview_from_memory():
    view = m.test_memoryview_from_memory()
    assert isinstance(view, memoryview)
    assert view.format == "B"
    assert bytes(view) == b"\xff\xe1\xab\x37"


def test_builtin_functions():
    assert m.get_len([i for i in range(42)]) == 42
    with pytest.raises(TypeError) as exc_info:
        m.get_len(i for i in range(42))
    assert str(exc_info.value) in [
        "object of type 'generator' has no len()",
        "'generator' has no length",
    ]  # PyPy


def test_isinstance_string_types():
    assert m.isinstance_pybind11_bytes(b"")
    assert not m.isinstance_pybind11_bytes(u"")

    assert m.isinstance_pybind11_str(u"")
    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE"):
        assert m.isinstance_pybind11_str(b"")
    else:
        assert not m.isinstance_pybind11_str(b"")


def test_pass_bytes_or_unicode_to_string_types():
    assert m.pass_to_pybind11_bytes(b"Bytes") == 5
    with pytest.raises(TypeError):
        m.pass_to_pybind11_bytes(u"Str")

    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE") or env.PY2:
        assert m.pass_to_pybind11_str(b"Bytes") == 5
    else:
        with pytest.raises(TypeError):
            m.pass_to_pybind11_str(b"Bytes")
    assert m.pass_to_pybind11_str(u"Str") == 3

    assert m.pass_to_std_string(b"Bytes") == 5
    assert m.pass_to_std_string(u"Str") == 3

    malformed_utf8 = b"\x80"
    if hasattr(m, "PYBIND11_STR_LEGACY_PERMISSIVE"):
        assert m.pass_to_pybind11_str(malformed_utf8) == 1
    elif env.PY2:
        with pytest.raises(UnicodeDecodeError):
            m.pass_to_pybind11_str(malformed_utf8)
    else:
        with pytest.raises(TypeError):
            m.pass_to_pybind11_str(malformed_utf8)


@pytest.mark.parametrize(
    "create_weakref, create_weakref_with_callback",
    [
        (m.weakref_from_handle, m.weakref_from_handle_and_function),
        (m.weakref_from_object, m.weakref_from_object_and_function),
    ],
)
def test_weakref(create_weakref, create_weakref_with_callback):
    from weakref import getweakrefcount

    # Apparently, you cannot weakly reference an object()
    class WeaklyReferenced(object):
        pass

    def callback(wr):
        # No `nonlocal` in Python 2
        callback.called = True

    obj = WeaklyReferenced()
    assert getweakrefcount(obj) == 0
    wr = create_weakref(obj)  # noqa: F841
    assert getweakrefcount(obj) == 1

    obj = WeaklyReferenced()
    assert getweakrefcount(obj) == 0
    callback.called = False
    wr = create_weakref_with_callback(obj, callback)  # noqa: F841
    assert getweakrefcount(obj) == 1
    assert not callback.called
    del obj
    pytest.gc_collect()
    assert callback.called



# Copyright (c) 2020 Jisang Yoon
# All rights reserved.
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.

# pylint: disable=no-name-in-module,too-few-public-methods,no-member
import os
import json
import tempfile

import numpy as np

from cuhnsw import aux
from cuhnsw.cuhnsw_bind import CuHNSWBind

EPS = 1e-10
WARP_SIZE = 32

class CuHNSW:
  def __init__(self, opt=None):
    self.opt = aux.get_opt_as_proto(opt or {})

    self.opt.level_mult = self.opt.level_mult or 1 / np.log(self.opt.max_m)
    self.logger = aux.get_logger("cuhnsw", self.opt.py_log_level)
    tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
    opt_content = json.dumps(aux.proto_to_dict(self.opt), indent=2)
    tmp.write(opt_content)
    tmp.close()
    self.logger.info("opt: %s", opt_content)
    self.data = None
    self.obj = CuHNSWBind()
    assert self.opt.block_dim <= WARP_SIZE ** 2 and \
      self.opt.block_dim % WARP_SIZE == 0, \
      f"invalid block dim ({self.opt.block_dim}, warp size: {WARP_SIZE})"
    assert self.obj.init(bytes(tmp.name, "utf8")), \
      f"failed to load {tmp.name}"
    os.remove(tmp.name)

  def set_data(self, data):
    self.data = data.copy()
    if self.opt.nrz:
      self.data /= np.linalg.norm(self.data, axis=1)[:, None]
    num_data, num_dims = self.data.shape
    self.logger.info("data shape: %d x %d", num_data, num_dims)
    self.obj.set_data(self.data)

  def build(self):
    self.set_random_levels()
    self.obj.build_graph()

  def set_random_levels(self):
    np.random.seed(self.opt.seed)
    num_data = self.data.shape[0]
    levels = np.random.uniform(size=num_data)
    levels = np.maximum(levels, EPS)
    levels = (-np.log(levels) * self.opt.level_mult).astype(np.int32)
    self.obj.set_random_levels(levels)

  def save_index(self, fpath):
    self.obj.save_index(fpath.encode("utf-8"))

  def load_index(self, fpath):
    self.obj.load_index(fpath.encode("utf-8"))

  def search_knn(self, qdata, topk, ef_search):
    ef_search = max(topk, ef_search)
    qdata = qdata.astype(np.float32)
    num_queries = qdata.shape[0]
    nns = np.empty(shape=(num_queries, topk), dtype=np.int32)
    distances = np.empty(shape=(num_queries, topk), dtype=np.float32)
    found_cnt = np.empty(shape=(num_queries,), dtype=np.int32)
    self.obj.search_knn(qdata, topk, ef_search,
                        nns, distances, found_cnt)
    return nns, distances, found_cnt



# Copyright (c) 2020 Jisang Yoon
# All rights reserved.
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.

# pylint: disable=fixme,too-few-public-methods
# reference: https://github.com/kakao/buffalo/blob/
# 5f571c2c7d8227e6625c6e538da929e4db11b66d/setup.py
"""cuhnsw
"""
import os
import sys
import glob
import pathlib
import platform
import sysconfig
import subprocess
from setuptools import setup, Extension

import numpy as np
from cuda_setup import CUDA, BUILDEXT


DOCLINES = __doc__.split("\n")

# TODO: Python3 Support
if sys.version_info[:3] < (3, 6):
  raise RuntimeError("Python version 3.6 or later required.")

assert platform.system() == 'Linux'  # TODO: MacOS


MAJOR = 0
MINOR = 0
MICRO = 6
RELEASE = True
STAGE = {True: '', False: 'b'}.get(RELEASE)
VERSION = f'{MAJOR}.{MINOR}.{MICRO}{STAGE}'
STATUS = {False: 'Development Status :: 4 - Beta',
          True: 'Development Status :: 5 - Production/Stable'}

CLASSIFIERS = """{status}
Programming Language :: C++
Programming Language :: Python :: 3.6
Operating System :: POSIX :: Linux
Operating System :: Unix
Operating System :: MacOS
License :: OSI Approved :: Apache Software License""".format( \
  status=STATUS.get(RELEASE))
CLIB_DIR = os.path.join(sysconfig.get_path('purelib'), 'cuhnsw')
LIBRARY_DIRS = [CLIB_DIR]
PYBIND_INCLUDE = "3rd/pybind11/include"

with open("requirements.txt", "r") as fin:
  INSTALL_REQUIRES = [line.strip() for line in fin]

def get_extend_compile_flags():
  flags = ['-march=native']
  return flags


class CMakeExtension(Extension):
  extension_type = 'cmake'

  def __init__(self, name):
    super().__init__(name, sources=[])


extend_compile_flags = get_extend_compile_flags()
extra_compile_args = ['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + \
  extend_compile_flags
csrcs = glob.glob("cpp/src/*.cu") + glob.glob("cpp/src/*.cc")
extensions = [
  # CMakeExtension(name="cuhnsw"),
  Extension("cuhnsw.cuhnsw_bind",
            sources= csrcs + [ \
              "cuhnsw/bindings.cc",
              "3rd/json11/json11.cpp"],
            language="c++",
            extra_compile_args=extra_compile_args,
            extra_link_args=["-fopenmp"],
            library_dirs=[CUDA['lib64']],
            libraries=['cudart', 'cublas', 'curand'],
            extra_objects=[],
            include_dirs=[ \
              "cpp/include/", np.get_include(),
              PYBIND_INCLUDE, CUDA['include'],
              "3rd/json11", "3rd/spdlog/include"])
]


# Return the git revision as a string
def git_version():
  def _minimal_ext_cmd(cmd):
    # construct minimal environment
    env = {}
    for k in ['SYSTEMROOT', 'PATH']:
      val = os.environ.get(k)
      if val is not None:
        env[k] = val
    out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env). \
      communicate()[0]
    return out

  try:
    out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
    git_revision = out.strip().decode('ascii')
  except OSError:
    git_revision = "Unknown"

  return git_revision


def write_version_py(filename='cuhnsw/version.py'):
  cnt = """
short_version = '%(version)s'
git_revision = '%(git_revision)s'
"""
  git_revision = git_version()
  with open(filename, 'w') as fout:
    fout.write(cnt % {'version': VERSION,
              'git_revision': git_revision})


class BuildExtension(BUILDEXT):
  def run(self):
    for ext in self.extensions:
      print(ext.name)
      if hasattr(ext, 'extension_type') and ext.extension_type == 'cmake':
        self.cmake()
    super().run()

  def cmake(self):
    cwd = pathlib.Path().absolute()

    build_temp = pathlib.Path(self.build_temp)
    build_temp.mkdir(parents=True, exist_ok=True)

    build_type = 'Debug' if self.debug else 'Release'

    cmake_args = [
      '-DCMAKE_BUILD_TYPE=' + build_type,
      '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + CLIB_DIR,
    ]

    build_args = []

    os.chdir(str(build_temp))
    self.spawn(['cmake', str(cwd)] + cmake_args)
    if not self.dry_run:
      self.spawn(['cmake', '--build', '.'] + build_args)
    os.chdir(str(cwd))


def setup_package():
  write_version_py()
  cmdclass = {
    'build_ext': BuildExtension
  }

  metadata = dict(
    name='cuhnsw',
    maintainer="Jisang Yoon",
    maintainer_email="vjs10101v@gmail.com",
    author="Jisang Yoon",
    author_email="vjs10101v@gmail.com",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    url="https://github.com/js1010/cuhnsw",
    download_url="https://github.com/js1010/cuhnsw/releases",
    include_package_data=False,
    license='Apache2',
    packages=['cuhnsw/'],
    cmdclass=cmdclass,
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    platforms=['Linux', 'Mac OSX', 'Unix'],
    ext_modules=extensions,
    install_requires=INSTALL_REQUIRES,
    entry_points={
      'console_scripts': [
      ]
    },
    python_requires='>=3.6',
  )

  metadata['version'] = VERSION
  setup(**metadata)


if __name__ == '__main__':
  setup_package()



# Copyright (c) 2020 Jisang Yoon
# All rights reserved.
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.

# Adapted from https://github.com/rmcgibbo/npcuda-example and
# https://github.com/cupy/cupy/blob/master/cupy_setup_build.py
# pylint: disable=fixme,access-member-before-definition
# pylint: disable=attribute-defined-outside-init,arguments-differ
import logging
import os
import sys

from distutils import ccompiler, errors, msvccompiler, unixccompiler
from setuptools.command.build_ext import build_ext as setuptools_build_ext

HALF_PRECISION = False

def find_in_path(name, path):
  "Find a file in a search path"
  # adapted fom http://code.activestate.com/
  # recipes/52224-find-a-file-given-a-search-path/
  for _dir in path.split(os.pathsep):
    binpath = os.path.join(_dir, name)
    if os.path.exists(binpath):
      return os.path.abspath(binpath)
  return None


def locate_cuda():
  """Locate the CUDA environment on the system
  If a valid cuda installation is found
  this returns a dict with keys 'home', 'nvcc', 'include',
  and 'lib64' and values giving the absolute path to each directory.
  Starts by looking for the CUDAHOME env variable.
  If not found, everything is based on finding
  'nvcc' in the PATH.
  If nvcc can't be found, this returns None
  """
  nvcc_bin = 'nvcc'
  if sys.platform.startswith("win"):
    nvcc_bin = 'nvcc.exe'

  # check env variables CUDA_HOME, CUDAHOME, CUDA_PATH.
  found = False
  for env_name in ['CUDA_PATH', 'CUDAHOME', 'CUDA_HOME']:
    if env_name not in os.environ:
      continue
    found = True
    home = os.environ[env_name]
    nvcc = os.path.join(home, 'bin', nvcc_bin)
    break
  if not found:
    # otherwise, search the PATH for NVCC
    nvcc = find_in_path(nvcc_bin, os.environ['PATH'])
    if nvcc is None:
      logging.warning('The nvcc binary could not be located in your '
              '$PATH. Either add it to '
              'your path, or set $CUDA_HOME to enable CUDA extensions')
      return None
    home = os.path.dirname(os.path.dirname(nvcc))

  cudaconfig = {'home': home,
          'nvcc': nvcc,
          'include': os.path.join(home, 'include'),
          'lib64':   os.path.join(home, 'lib64')}
  post_args = [
    "-arch=sm_52",
    "-gencode=arch=compute_52,code=sm_52",
    "-gencode=arch=compute_60,code=sm_60",
    "-gencode=arch=compute_61,code=sm_61",
    "-gencode=arch=compute_70,code=sm_70",
    "-gencode=arch=compute_75,code=sm_75",
    "-gencode=arch=compute_80,code=sm_80",
    "-gencode=arch=compute_86,code=sm_86",
    "-gencode=arch=compute_86,code=compute_86",
    '--ptxas-options=-v', '-O2']
  if HALF_PRECISION:
    post_args = [flag for flag in post_args if "52" not in flag]

  if sys.platform == "win32":
    cudaconfig['lib64'] = os.path.join(home, 'lib', 'x64')
    post_args += ['-Xcompiler', '/MD', '-std=c++14',  "-Xcompiler", "/openmp"]
    if HALF_PRECISION:
      post_args += ["-Xcompiler", "/D HALF_PRECISION"]
  else:
    post_args += ['-c', '--compiler-options', "'-fPIC'",
                  "--compiler-options", "'-std=c++14'"]
    if HALF_PRECISION:
      post_args += ["--compiler-options", "'-D HALF_PRECISION'"]
  for k, val in cudaconfig.items():
    if not os.path.exists(val):
      logging.warning('The CUDA %s path could not be located in %s', k, val)
      return None

  cudaconfig['post_args'] = post_args
  return cudaconfig


# This code to build .cu extensions with nvcc is taken from cupy:
# https://github.com/cupy/cupy/blob/master/cupy_setup_build.py
class _UnixCCompiler(unixccompiler.UnixCCompiler):
  src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
  src_extensions.append('.cu')

  def _compile(self, obj, src, ext, cc_args, extra_postargs, pp_opts):
    # For sources other than CUDA C ones, just call the super class method.
    if os.path.splitext(src)[1] != '.cu':
      return unixccompiler.UnixCCompiler._compile(
        self, obj, src, ext, cc_args, extra_postargs, pp_opts)

    # For CUDA C source files, compile them with NVCC.
    _compiler_so = self.compiler_so
    try:
      nvcc_path = CUDA['nvcc']
      post_args = CUDA['post_args']
      # TODO? base_opts = build.get_compiler_base_options()
      self.set_executable('compiler_so', nvcc_path)

      return unixccompiler.UnixCCompiler._compile(
        self, obj, src, ext, cc_args, post_args, pp_opts)
    finally:
      self.compiler_so = _compiler_so


class _MSVCCompiler(msvccompiler.MSVCCompiler):
  _cu_extensions = ['.cu']

  src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
  src_extensions.extend(_cu_extensions)

  def _compile_cu(self, sources, output_dir=None, macros=None,
          include_dirs=None, debug=0, extra_preargs=None,
          extra_postargs=None, depends=None):
    # Compile CUDA C files, mainly derived from UnixCCompiler._compile().
    macros, objects, extra_postargs, pp_opts, _build = \
      self._setup_compile(output_dir, macros, include_dirs, sources,
                depends, extra_postargs)

    compiler_so = CUDA['nvcc']
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
    post_args = CUDA['post_args']

    for obj in objects:
      try:
        src, _ = _build[obj]
      except KeyError:
        continue
      try:
        self.spawn([compiler_so] + cc_args + [src, '-o', obj] + post_args)
      except errors.DistutilsExecError as e:
        raise errors.CompileError(str(e))

    return objects

  def compile(self, sources, **kwargs):
    # Split CUDA C sources and others.
    cu_sources = []
    other_sources = []
    for source in sources:
      if os.path.splitext(source)[1] == '.cu':
        cu_sources.append(source)
      else:
        other_sources.append(source)

    # Compile source files other than CUDA C ones.
    other_objects = msvccompiler.MSVCCompiler.compile(
      self, other_sources, **kwargs)

    # Compile CUDA C sources.
    cu_objects = self._compile_cu(cu_sources, **kwargs)

    # Return compiled object filenames.
    return other_objects + cu_objects


class CudaBuildExt(setuptools_build_ext):
  """Custom `build_ext` command to include CUDA C source files."""

  def run(self):
    if CUDA is not None:
      def wrap_new_compiler(func):
        def _wrap_new_compiler(*args, **kwargs):
          try:
            return func(*args, **kwargs)
          except errors.DistutilsPlatformError:
            if sys.platform != 'win32':
              CCompiler = _UnixCCompiler
            else:
              CCompiler = _MSVCCompiler
            return CCompiler(
              None, kwargs['dry_run'], kwargs['force'])
        return _wrap_new_compiler
      ccompiler.new_compiler = wrap_new_compiler(ccompiler.new_compiler)
      # Intentionally causes DistutilsPlatformError in
      # ccompiler.new_compiler() function to hook.
      self.compiler = 'nvidia'

    setuptools_build_ext.run(self)


CUDA = locate_cuda()
assert CUDA is not None
BUILDEXT = CudaBuildExt if CUDA else setuptools_build_ext
